/* ==========================================================================*/
/* - - - - - - - - - -  P R O C S I G N A L , H  - - - - - - - - - - - - - - */
/* ==========================================================================*/
/*	Created   07-10-2004   M. Wheatley                                       */
/*	Modified  08-09-2004   MW                                                */
/*.........................................................................  */
/* PROCSIGNAL Module  File                                                   */
/*.........................................................................  */
/* Processing Module                                                         */
/*.........................................................................  */

/*---------------------------------------------------------------------------*/
/*------------------------> I N C L U D E S <--------------------------------*/
/*---------------------------------------------------------------------------*/
#include <p30f6014.h>
#include <dsp.h>
#include "common.h"
#include "codec.h"
#include "menu.h"
#include "timers.h"
#include "gettime.h"
#include "tables.h"

/*---------------------------------------------------------------------------*/
/*----------------> P R O C E S S O R    S P E C I F I C <-------------------*/
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/*--------> T Y P E D E F S   U N I O N S    S T R U C T U R E S <-----------*/
/*---------------------------------------------------------------------------*/

/*---------------------------------------------------------------------------*/
/*--------------------> L O C A L   D E F I N E S <--------------------------*/
/*---------------------------------------------------------------------------*/
#define SECSYNCBUF_LENGTH 150
#define RAWDATABUF_LENGTH (150/5)
#define SYNDATA_FILTER_DELAY 21	//delay of filters + 30mSec hole before data
#define DATA_FILTER_DELAY 21	//delay of filters + 30mSec hole before data
#define TIC_QUALITY_LIM1 5		//Limit of consecutive sec position finds
#define TIC_QUALITY_LIM2 10		//	before saying it is a good position
#define MIN_QUALITY_LIM 5

#define MAX_LIMIT 500

/*---------------------------------------------------------------------------*/
/*----------------> E X T E R N A L  R E F E R E N C E S <-------------------*/
/*---------------------------------------------------------------------------*/
//asm helper functions (defined in AsmUtils.s)
extern void CalcABS( unsigned int Size, fractional* pIn);
extern void CalcLpIir( unsigned int Size, int* pIn,
						 fractional* pLPVal, unsigned int LPK); 

/*---------------------------------------------------------------------------*/
/*--------------> G L O B A L  M E M O R Y  A L L O C A T I O N <------------*/
/*---------------------------------------------------------------------------*/
int TicPos;
int SecSyncBuf[SECSYNCBUF_LENGTH];	//expose for display plotting
int RawDataBuf[RAWDATABUF_LENGTH+1];
char MinSyncQuality;

/*---------------------------------------------------------------------------*/
/*--------------> L O C A L  M E M O R Y  A L L O C A T I O N <--------------*/
/*---------------------------------------------------------------------------*/
static unsigned int page;
static int TicLastPos;
static int TicQuality;
static int SecSyncPtr;
static int MinSyncave;
static int Dec2Buf[PROC_BLOCK_SIZE/2];	//used by 1KHz and 100Hz data streams
static int Buf100[PROC_BLOCK_SIZE/(2*4)];	//used by 100Hz data stream
static unsigned long MinuteSyncSignal;
static unsigned long MinuteSyncNoise;
static fractional* offset;
static fractional EarlyAve;
static fractional LateAve;
static FIRStruct fDec2LP1800;		//FIR to decimate from 7200 to 3600 sps
static FIRStruct fBPTic;			//FIR to BP filter the TIC 1 or 1.2 KHz tone
static FIRStruct fDec4LP200;		//FIR to decimate the TIC pulse to 900sps
static FIRStruct fDec4LP110;		//FIR to decimate the 100Hz data 3600 to 900 sps
static FIRStruct fDec2LP110;		//FIR to decimate the 100Hz data 900 to 450 sps
static FIRStruct fBP100;			//FIR to BP filter the 100Hz data tone

static fractional DelayDec2LP1800[DEC2LP1800_SIZE] __attribute__ ( (__section__(".ybss")) );
static fractional DelayBPTic[BPTIC_SIZE] __attribute__ ( (__section__(".ybss")) );
static fractional DelayDec4LP200[DEC4LP200_SIZE] __attribute__ ( (__section__(".ybss")) );
static fractional DelayDec4LP110[DEC4LP110_SIZE] __attribute__ ( (__section__(".ybss")) );
static fractional DelayDec2LP110[DEC2LP110_SIZE] __attribute__ ( (__section__(".ybss")) );
static fractional DelayBP100[BP100_SIZE] __attribute__ ( (__section__(".ybss")) );

/*---------------------------------------------------------------------------*/
/*-----------------------> P R O T O T Y P E S <-----------------------------*/
/*---------------------------------------------------------------------------*/
void InitProcSignal( void );
void ProcInput( fractional* pIn);
void FindSecSync(void);
void FindMinSync(void);
char CalcBit(void );
static void CalcClockError(void);

/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
/*                       -----------------------------                       */
/*                      | I n i t P r o c S i g n a l |                      */
/*                       -----------------------------                       */
/*    This initializes the Signal Processing module.                         */
/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
void InitProcSignal( void)
{
int i;
	asm("mov #psvpage(_DEC2LP1800),w0\nmov w0,_page");
	asm("mov #psvoffset(_DEC2LP1800),w0\nmov w0,_offset");
	FIRStructInit(&fDec2LP1800,DEC2LP1800_SIZE, offset, page, DelayDec2LP1800);
 	FIRDelayInit(&fDec2LP1800);

	if(SemB.WWV)
	{
		asm("mov #psvpage(_BPTIC1000),w0\nmov w0,_page");
		asm("mov #psvoffset(_BPTIC1000),w0\nmov w0,_offset");
	}
	else
	{
		asm("mov #psvpage(_BPTIC1200),w0\nmov w0,_page");
		asm("mov #psvoffset(_BPTIC1200),w0\nmov w0,_offset");
	}
	FIRStructInit(&fBPTic, BPTIC_SIZE, offset, page, DelayBPTic);
 	FIRDelayInit(&fBPTic);

	asm("mov #psvpage(_DEC4LP200),w0\nmov w0,_page");
	asm("mov #psvoffset(_DEC4LP200),w0\nmov w0,_offset");
	FIRStructInit(&fDec4LP200,DEC4LP200_SIZE, offset, page, DelayDec4LP200);
 	FIRDelayInit(&fDec4LP200);

	asm("mov #psvpage(_DEC4LP110),w0\nmov w0,_page");
	asm("mov #psvoffset(_DEC4LP110),w0\nmov w0,_offset");
	FIRStructInit(&fDec4LP110,DEC4LP110_SIZE, offset, page, DelayDec4LP110);
 	FIRDelayInit(&fDec4LP110);

	asm("mov #psvpage(_DEC2LP110),w0\nmov w0,_page");
	asm("mov #psvoffset(_DEC2LP110),w0\nmov w0,_offset");
	FIRStructInit(&fDec2LP110,DEC2LP110_SIZE, offset, page, DelayDec2LP110);
 	FIRDelayInit(&fDec2LP110);

	asm("mov #psvpage(_BP100),w0\nmov w0,_page");
	asm("mov #psvoffset(_BP100),w0\nmov w0,_offset");
	FIRStructInit(&fBP100, BP100_SIZE, offset, page, DelayBP100);
 	FIRDelayInit(&fBP100);

	TicPos = 0;
	SecSyncPtr = 0;
	for(i=0; i<SECSYNCBUF_LENGTH; i++)
		SecSyncBuf[i] = 0;
	for(i=0; i<RAWDATABUF_LENGTH+1; i++)
		RawDataBuf[i] = 0;
	SemA.SecLocked = FALSE;
	SemA.MinLocked = FALSE;
	EarlyAve = 0;
	LateAve = 0;
	MinSyncQuality = 0;
	MinuteSyncSignal = 0;
	MinuteSyncNoise = 0;
}


/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
/*                          -------------------                              */
/*                         | P r o c I n p u t |                             */
/*                          -------------------                              */
/*    This processes the input stream and creates a seconds/minute sync      */
/*  buffer and a data bit buffer. Called with PROC_BLOCK_SIZE (20mSec) worth */
/*  of data to be processed.                                                 */
/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
void ProcInput( fractional* pIn)
{
int i;
int j;
	//decimate the 7200 sps stream to 3600 sps
	FIRDecimate( PROC_BLOCK_SIZE/2, Dec2Buf, pIn, &fDec2LP1800, 2);

	//decimate the 3600 sps stream to 900 sps for the 100Hz data path
	FIRDecimate( PROC_BLOCK_SIZE/(2*4), Buf100, Dec2Buf, &fDec4LP110, 4);

	// BP Filter the 3600 sps stream to get the 1KHz Tics
	FIR( PROC_BLOCK_SIZE/2, Dec2Buf, Dec2Buf, &fBPTic);

	// "rectify" to demod the the 1KHz Tics.
	CalcABS( PROC_BLOCK_SIZE/2, Dec2Buf);	//

	//decimate the demod stream to 900 sps
	FIRDecimate( PROC_BLOCK_SIZE/(2*4), Dec2Buf, Dec2Buf, &fDec4LP200, 4);

	if(SemA.SecLocked)	//if seconds are locked then get minute sync energy
	{
		if(SecSyncPtr > 3 )	//skip TIC energy bins
		{
			if(SecSyncPtr < (SECSYNCBUF_LENGTH*8)/10 )
			{	//average energy in first 800mS of seconds sync buffer
				for(i=0; i<PROC_BLOCK_SIZE/(2*4); i++)
					MinuteSyncSignal = MinuteSyncSignal + (unsigned long)Dec2Buf[i]; 
			}
			else
			{	//average energy in last 200mS of seconds sync buffer
				for(i=0; i<PROC_BLOCK_SIZE/(2*4); i++)
					MinuteSyncNoise = MinuteSyncNoise + (unsigned long)Dec2Buf[i]; 
			}
		}
	}
	else
	{
		MinuteSyncSignal = 0;
		MinuteSyncNoise = 0;
	}
	// LP each individual time bin of the "rectified" 1KHz Tic pulses
	// and place in SecSyncBuf. Takes PROC_BLOCK_SIZE/(2*4) samples and
	// places them in 3 SecSyncBuf bins, averaging them with an IIR filter.
	CalcLpIir( PROC_BLOCK_SIZE/(2*4*3), &Dec2Buf[0], &SecSyncBuf[SecSyncPtr], 1000);
	SecSyncPtr++;
	if(SecSyncPtr>=SECSYNCBUF_LENGTH)	//deal with buffer wrap
	{
		SecSyncPtr = 0;
		SemA.SecRdy = TRUE;
	}
	CalcLpIir( PROC_BLOCK_SIZE/(2*4*3), &Dec2Buf[6], &SecSyncBuf[SecSyncPtr], 1000); 
	SecSyncPtr++;
	if(SecSyncPtr>=SECSYNCBUF_LENGTH)	//deal with buffer wrap
	{
		SecSyncPtr = 0;
		SemA.SecRdy = TRUE;
	}
	CalcLpIir( PROC_BLOCK_SIZE/(2*4*3), &Dec2Buf[12], &SecSyncBuf[SecSyncPtr], 1000); 
	SecSyncPtr++;
	if(SecSyncPtr>=SECSYNCBUF_LENGTH)	//deal with buffer wrap
	{
		SecSyncPtr = 0;
		SemA.SecRdy = TRUE;
	}

	if(SemA.MinLocked)	//if seconds and minutes are locked then get data bits
	{
		//	Now process the 100Hz data stream for data bit extraction
		//decimate the 900 sps stream to 450 sps for the 100Hz data path
		FIRDecimate( PROC_BLOCK_SIZE/(2*4*2), Buf100, Buf100, &fDec2LP110, 2);
		// BP Filter the 450 sps stream to get the 100Hz signal
		FIR( PROC_BLOCK_SIZE/(2*4*2), Buf100, Buf100, &fBP100);
		// "rectify" to demod the the 100Hz Tone.
		CalcABS( PROC_BLOCK_SIZE/(2*4*2), Buf100);
		if( SecSyncPtr > DATA_FILTER_DELAY )
			i = SecSyncPtr - DATA_FILTER_DELAY;
		else
			i = SECSYNCBUF_LENGTH - (DATA_FILTER_DELAY - SecSyncPtr);
		i = i / (SECSYNCBUF_LENGTH/RAWDATABUF_LENGTH);
		CalcLpIir( PROC_BLOCK_SIZE/(2*4*2), &Buf100[0], &RawDataBuf[i], 20000);
	}
}

/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
/*                          ---------------                                  */
/*                         | C a l c B i t |                                 */
/*                          ---------------                                  */
/*  Finds the current data bit value by comparing zero and one signal levels */
/* to the noise level and slicing it into a ternary value					 */
/*  +1 == ONE    -1 == ZERO      0 == Not determined						 */
/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
char CalcBit(void )
{
int i;
int noise;
int signal0;
int signal1;
	noise = 0;
	signal0 = 0;
	signal1 = 0;
	for(i=0; i<=4; i++)				//integrate ZERO bit time slot
		signal0 += RawDataBuf[i];
	for(i=7; i<=11; i++)			//integrate ONE bit time slot
		signal1 += RawDataBuf[i];
	for(i=24; i<=28; i++)
		noise += RawDataBuf[i];		//integrate Noise time slot
	noise = noise << 2;
	if(signal1 > noise)				// need S/N > 4
		return 1;
	else
		if(signal0 > noise)
			return -1;
		else
			return 0;
}


/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
/*                          -----------------------                          */
/*                         | F i n d S e c S y n c |                         */
/*                          -----------------------                          */
/*  Finds the Seconds position by finding the energy peak in the 1 or 1.2KHz */
/*   circular energy buffer.  Each bin is 1/150 seconds.                     */
/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
void FindSecSync(void)
{
int max;
	// find peak position in SecSyncBuf
	max = VectorMax(SECSYNCBUF_LENGTH, SecSyncBuf, &TicPos);
	if( max < MAX_LIMIT )
	{
		SemA.SecLocked = FALSE;
		SemA.MinLocked = FALSE;
		MinSyncQuality = 0;
		EarlyAve = 0;
		LateAve = 0;
		LED2_OFF
		LED3_OFF
		return;
	}

	if(!SemA.SecLocked)
	{
		if( TicPos == TicLastPos )
		{
			TicQuality++;
			if(TicQuality>TIC_QUALITY_LIM1)	//clamp to Limit
				TicQuality = TIC_QUALITY_LIM1;
		}
		else
		{
			TicQuality = 0;
		}
		if(TicQuality ==TIC_QUALITY_LIM1)
		{
			VectorScale(SECSYNCBUF_LENGTH, SecSyncBuf, SecSyncBuf, 0);	//clear array
			SecSyncBuf[0] = max;	//restore peak power to zero position
			if(SecSyncPtr >= TicPos )
				SecSyncPtr = SecSyncPtr - TicPos;
			else
				SecSyncPtr = SECSYNCBUF_LENGTH - (TicPos - SecSyncPtr);
			TicPos = 0;
			SemA.SecLocked = TRUE;
			LED2_ON
		}
	}
	else
	{
		if( 0 == TicPos )
		{
			TicQuality++;
			if(TicQuality>TIC_QUALITY_LIM2 )	//clamp to Limit
				TicQuality = TIC_QUALITY_LIM2;
			if(SemA.MinLocked)
				CalcClockError();
		}
		else
		{
			TicQuality--;
			if( TicQuality<=0 )
			{
				VectorScale(SECSYNCBUF_LENGTH, SecSyncBuf, SecSyncBuf, 0);	//clear array
				SemA.SecLocked = FALSE;
				SemA.MinLocked = FALSE;
				MinSyncQuality = 0;
				EarlyAve = 0;
				LateAve = 0;
				LED2_OFF
				LED3_OFF
			}
		}
	}
	TicLastPos = TicPos;
}

/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
/*                          -----------------------                          */
/*                         | F i n d M i n S y n c |                         */
/*                          -----------------------                          */
/*  Finds the Minutes position by looking at the energy peak for the 800mS   */
/*   Pulse at the beginning of each minute.                                  */
/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
void FindMinSync(void)
{
	if( MinuteSyncSignal > (MinuteSyncNoise*4*5) )
	{	//if got a minutesync signal
		if(!SemA.MinLocked)
		{
			SemA.MinLocked = TRUE;
			SecondsCount = 0;
			MinSyncQuality = 1;
			LED3_ON
		}
		else
		{
			if( 0 == SecondsCount)
				MinSyncQuality++;
			else
				MinSyncQuality--;
			if(MinSyncQuality>MIN_QUALITY_LIM)
				MinSyncQuality = MIN_QUALITY_LIM;
			if(MinSyncQuality<0)
			{
				MinSyncQuality = 0;
				SemA.MinLocked = FALSE;
				EarlyAve = 0;
				LateAve = 0;
				LED3_OFF
			}
		}
	}
	MinuteSyncSignal = 0;
	MinuteSyncNoise = 0;
}

/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
/*                     -----------------------------                         */
/*                    | C a l c C l o c k E r r o r |                        */
/*                     -----------------------------                         */
/*  Try to keep seconds sync puls centered in the Zero bin of SecSyncBuf     */
/*%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%*/
static void CalcClockError(void)
{
	//average bin before and after sync pulse
	CalcLpIir( 1, &SecSyncBuf[149], &EarlyAve, 1000);
	CalcLpIir( 1, &SecSyncBuf[1], &LateAve, 1000);
	if( (SecondsCount & 0x07) == 0 )	//every 8 seconds
	{
		if(LateAve > (EarlyAve+400) )
			AdjustClock( -1 );			//slip one codec data sample
		if( EarlyAve > (LateAve+400) )
			AdjustClock( 1 );			//add one codec data sample
	}
	//to do...could calculate clock error in ppm and save in EEROM
}
